#ifndef INCLUDED_AGENT_COMMAND_H
#define INCLUDED_AGENT_COMMAND_H

#include <string>
#include <list>

#include "command.h"

class SIOAgentProxy;
struct Tcl_Interp;

/*
    An Agent Command is a command that requires a locked agent list in order
    for the command to work
*/

class AgentCommand {
public:
    virtual ~AgentCommand();

    /*
        This is the virtual function that you override that requires the 
        locked context
    */
    virtual void Execute(SIOAgentProxy* agent) = 0;
};

/*
    This class encapsulates adding a data wme
*/
class AddDataWMECommand : public AgentCommand {
public:
     AddDataWMECommand(int inParentId,const char* inAttribute,
                       const char* inValue,int inSimTimeTag)
     : parentId(inParentId),attr(inAttribute),
       value(inValue), simTimeTag(inSimTimeTag) {}

     virtual ~AddDataWMECommand();

     virtual void Execute(SIOAgentProxy* agent);
private:
     int         parentId;
     std::string attr;
     std::string value;
     int         simTimeTag;
};

/*
    This class encapsulates adding a plain wme
*/
class AddPlainWMECommand : public AgentCommand {
public:
     AddPlainWMECommand(const char* inParent,const char* inAttribute,
                        const char* inValue)
     : parent(inParent),attr(inAttribute),value(inValue) {}

     virtual ~AddPlainWMECommand();

     virtual void Execute(SIOAgentProxy* agent);
private:
     std::string parent;
     std::string attr;
     std::string value;
};

/*
    This class encapsulates adding a wme whose value will be an id
*/
class AddIdWMECommand 
     : public AgentCommand {
public:
     AddIdWMECommand(int inParentId,const char* inAttribute,
                     int inChildId,int inSimTimeTag)
     : parentId(inParentId),attr(inAttribute),
       childId(inChildId), simTimeTag(inSimTimeTag) {}

     virtual ~AddIdWMECommand();

     virtual void Execute(SIOAgentProxy* agent);
private:
     int         parentId;
     std::string attr;
     int         childId;
     int         simTimeTag;
};

/*
    This class encapsulates removing a wme
*/
class RemoveWMECommand 
     : public AgentCommand {
public:
     RemoveWMECommand(int inSimTimeTag) 
     : simTimeTag(inSimTimeTag) {}

     virtual ~RemoveWMECommand();

     virtual void Execute(SIOAgentProxy* agent);
private:
     int simTimeTag;
};

/*
    This class encapsulates commiting working memory changes
*/
class CommitCommand 
     : public AgentCommand {
public:
     virtual ~CommitCommand();

     virtual void Execute(SIOAgentProxy* agent);
};

/*
    This class encapsulates commiting working memory changes
*/
class DoWMEChangesCommand 
     : public AgentCommand {
public:
     DoWMEChangesCommand(struct Tcl_Interp* inInterp)
          : interp(inInterp) {}

     virtual ~DoWMEChangesCommand();

     virtual void Execute(SIOAgentProxy* agent);
private:
     Tcl_Interp* interp;
};

/*
    This class encapsulates reporting commands to the soarside
    during the output-link phase
*/
class ReportCommands 
     : public AgentCommand {
public:
     ReportCommands(const std::list<Command>& inCmds)
          : cmds(inCmds) {}

     virtual ~ReportCommands();

     virtual void Execute(SIOAgentProxy* agent);
private:
     const std::list<Command>& cmds;
};

#endif // INCLUDED_AGENT_COMMAND_H