#ifndef INCLUDED_SENDER_H
#define INCLUDED_SENDER_H


#include "sync_queue.h"
#include "thread.h"

namespace sgio {
	class Socket;
	class Message;
}

/**
   Sender is a class in which messages are posted, and sent over
   the socket from a seperate thread so the calling thread does not
   have to block on it

   example of how this code works
   Sender* sender = new Sender(socket);

   // Post some messages
   ....
   Message* message  = new AddNamedAgentMessage("bob");
   sender->PostNetMessage(message);

   // when you are done
   delete Sender;
 */
class Sender {
public:
   /**
      This constructs a sender object attached to the passed in socket
      the reference count is set to 1 at this point
      @param inSocket the socket you want this sender attached to
   */
	Sender(sgio::Socket* inSocket);

     ~Sender();

	/**
      This method posts a message that is supposed to be sent over the socket, 
      it must be dynamically allocated with new, after this call you no longer 
      own the memory, do not delete it
      @param inMessage the message to be sent over the network
   */
     void PostNetMessage(sgio::Message* inMessage) {
          d_msgQueue.Enqueue(inMessage);
     }
	
private:
	// NOT IMPLEMENTED
	Sender();
	Sender(const Sender&);
	Sender& operator=(const Sender&);

	// Thread Functions
	static void thread_func(void* self);
	void ThreadFunc();

     sgio::Message* GetNetMessage() {
          return d_msgQueue.Dequeue();
     }
 
     SyncQueue<sgio::Message*> d_msgQueue;
   
	
     sgio::Socket* d_sock;
     Thread::SingleEvent d_quitSignal;

};


#endif // INCLUDED_SENDER_H 
