#ifndef SIO_PROXY_AGENT_H
#define SIO_PROXY_AGENT_H

#pragma warning(disable:4786)

#include <queue>
#include <list>
#include <string>
#include <map>
#include <set>
#include <algorithm>
#include "sgio_frame.h"
#include "thread.h"

struct Tcl_Interp;

class Command;
class Sender;

/**
    This class is a slave to the SIO Agent on the simside,
    it keeps the translation data from simside id's to
    soar Id's
*/

class SIOAgentProxy : public sgio::WMChangeExecuter {
public:
    /*
        Creates an agent proxy 
        
        @param inName the name of this agent
        @param inSender where feedback messages should be sent
    */
	SIOAgentProxy(const char* inName,Sender* inSender);

	~SIOAgentProxy();

    /*
        This call corresponds to an agent->AddWME call on the simside
        that is the client calls AddWME on an SIO agent on the simside
        then this call will eventually be made on this side of the socket

        @param simsideId the id that the simside is using for the parent

        @param attribute the name of the attribute for this wme

        @param childSimsideId the id the simside assigned to the new 
               soarId that will be created
        
        @param simsideTimeTag the id that the simside uses to keep track of
               this WME
    */
	void AddWME(int simsideId,const std::string& attribute,int childSimsideId,int simsideTimeTag);

    /**
        @param simsideId the id that the simside is using for the parent

        @param attribute the name of the attribute for this wme

        @param value the value for this wme

        @param simsideTimeTag the id that the simside uses to keep track of this
               WME
    */
	void AddWME(int simsideId,const std::string& attribute,const std::string& value,int simsideTimeTag);

    /**
        @param soarId the soarId parent for where this wme should be added

        @param attribute the attribute for this wme

        @param value the value for this wme
    */
    void AddWME(const std::string& soarId,const std::string& attribute,const std::string& value);

    /*
        This removes a wme from working memory
        this call corresponds to a agent->RemoveWME call on the simside

        @param simsideTimeTag the id that the simside used to keep track of this
               WME that was created with a previous AddWME call
    */
	void RemoveWME(int simsideTimeTag);

    /*
        This states that a consistent state of the world has been reached
        and can be presented to soar as soon as it is ready
    */
    void Commit();
	
    /*
        This sets up context for this object because
        DoWMChanges cannot take any parameters
    */
    void SetTclInterp(Tcl_Interp* inInterp) { interp = inInterp; }

    /*
        This clears the context for DoWMChanges
    */
    void ClearTclInterp() { interp = 0; }

    /*
        soar is ready for working memory changes, do any changes queued
        up
    */
	void DoWMChanges();

    /*
        This is part of the WMChangeExecuter interface so that we interact
        with the sgio::Frame class, corresponding lookups are mapped and
        actual changes to working memory take place
    */
	virtual void DoAddWME(int inParentId,const std::string& inAttribute,int inId,int inWmeId);
	virtual void DoAddWME(int simsideId,const std::string& attribute,const std::string& value,int simsideTimeTag);
    virtual void DoAddWME(const std::string& soarId,const std::string& attribute,const std::string& value);
	virtual void DoRemoveWME(int simsideId);

    /* 
        Gets the name of this agent
    */
	const std::string& GetName() const { return name; }
	
    /*
        This translates commands into command messages and
        sends them over the socket

        @param inCommand the command that soar reported to us
    */
	void AddCommand(const Command& inCommand);

    /*
        Soar reports the same commands over and over again, but we don't want
        to repeated send over commands we are processing on the simside, so
        keep track of what commands we have seen this time and last time, in
        order to properly do this we must know when the ouput-phase ends

        this function call notifies this object when this happens.
    */
    void EndCommands();
private:
    // NOT implemented
	SIOAgentProxy();
	SIOAgentProxy(const SIOAgentProxy&);
	SIOAgentProxy& operator=(const SIOAgentProxy&);


    Sender* const sender;

	std::string name;

    
	sgio::Frame building;

	Thread::Mutex completeLock;
    sgio::Frame complete;

	typedef std::map<int,std::string>::iterator misIter;
	std::map<int,std::string> idToSoarId;

	typedef std::map<int,int>::iterator miiIter;
	std::map<int,int> wmeIdToTimeTag; 

    std::map<int,int> wmeIdToId;

    std::set<std::string> commandsSeenThisTime;
    std::set<std::string> commandsSeenLastTime;
    Tcl_Interp* interp;
};

#endif // SIO_PROXY_AGENT_H
