#pragma warning(disable:4786)

#include "sio_soar_proxy.h"
#include "sender.h"
#include "receiver.h"
#include "tcl_commands.h"
#include "agent_commands.h"
#include "sio_agent_proxy.h"
#include <functional>
#include <algorithm>

using namespace std;

struct AgentNameEquals
: public std::binary_function<const SIOAgentProxy*,std::string,bool> {
    bool operator()(const SIOAgentProxy* agentPtr,const std::string& name) const {
        return agentPtr->GetName() == name;
    }

};



SIOSoarProxy::SIOSoarProxy(std::auto_ptr<sgio::Socket> client) 
: d_socket(client), 
  d_sender(new Sender(d_socket.get())),
  quit(false) {
      d_receiver = std::auto_ptr<Receiver>(new Receiver(this,d_socket.get(),d_sender.get()));
}

SIOSoarProxy::~SIOSoarProxy() {
	{
		Thread::Monitor m(&commandsLock);
		while(!commands.empty()) {
			delete commands.front();
			commands.pop_front();
		}

	}
    {
        Thread::Monitor m(&d_agentsListLock);
        typedef std::list<SIOAgentProxy*>::iterator iter;

        iter i = d_agentsList.begin();
        while(i != d_agentsList.end()) {
            SIOAgentProxy* currAgent = *i;
            delete currAgent;
            i = d_agentsList.erase(i);
        }
    }
}

void 
SIOSoarProxy::PostNetMessage(std::auto_ptr<sgio::Message> msg) {
     d_sender->PostNetMessage(msg.release());
}


void
SIOSoarProxy::RunAgentCommand(const std::string& agentName,AgentCommand& agentCmd) {
    // Lock the list, there might be contension for the list if
    // if soar is currently thinking
    Thread::Monitor m(&d_agentsListLock);

    // find the agent with the same name that is passed in
    AgentsList::iterator i = std::find_if(d_agentsList.begin(),d_agentsList.end(),
                                          bind2nd(AgentNameEquals(),agentName));
    
    // if we found the agent, execute the command
    if(i != d_agentsList.end()) {
        agentCmd.Execute(*i);
    }
}


void 
SIOSoarProxy::CreateAgent(const std::string& inAgentName) {
    // Tell soar to create the agent eventually
    std::auto_ptr<TclCommand> addAgentCmd(new AddAgentTclCmd(inAgentName));
    PostTclCommand(addAgentCmd);

    // Create the data structures that we need for book-keeping
    Thread::Monitor m(&d_agentsListLock);
    d_agentsList.push_back(new SIOAgentProxy(inAgentName.c_str(),d_sender.get()));
}

void
SIOSoarProxy::DestroyAgent(const std::string& inAgentName) {
    Thread::Monitor m(&d_agentsListLock);

    AgentsList::iterator i = std::find_if(d_agentsList.begin(),d_agentsList.end(),
                                          bind2nd(AgentNameEquals(),inAgentName));

    if(i != d_agentsList.end()) {
        delete *i;
        d_agentsList.erase(i);
    }
}

std::string
SIOSoarProxy::GetAnyAgentName() {
     Thread::Monitor m(&d_agentsListLock);
     if(!d_agentsList.empty())
          return d_agentsList.front()->GetName();
     else
          return std::string();
}

void 
SIOSoarProxy::PostTclCommand(std::auto_ptr<TclCommand> inCommand) {
    Thread::Monitor m(&commandsLock);
	if(inCommand->Unique()) {
         /*
            The expression in the conditional
            sees if for any TclCommand* x in the list
            x->Obliterates(inCommand) returns true
         */
         if(find_if(commands.begin(),commands.end(),
                    bind2nd(mem_fun1(TclCommand::Obliterates),inCommand.get())) != commands.end()) {
            return;
		 }
	}
	commands.push_back(inCommand.release());
}

std::auto_ptr<TclCommand>
SIOSoarProxy::GetTclCommand() {
    Thread::Monitor m(&commandsLock);
	TclCommand* tclCommand = commands.front();
	commands.pop_front();
	return std::auto_ptr<TclCommand>(tclCommand);
}

bool 
SIOSoarProxy::IsATclCommandPending() {
    Thread::Monitor m(&commandsLock);
    return !commands.empty();
}