#ifndef INCLUDED_SIO_SOAR_PROXY_H
#define INCLUDED_SIO_SOAR_PROXY_H

class AgentCommand;
class TclCommand;
class Sender;
class Receiver;
class SIOAgentProxy;

// includes
#ifndef INCLUDED_THREAD_H
#include "thread.h"
#endif

#ifndef INCLUDED_SGIO_SOCKET_H
#include "sgio_socket.h"
#endif

#ifndef INCLUDED_SGIO_MESSAGE_H
#include "sgio_messages.h"
#endif


#include <string>
#include <list>
#include <memory>

/*
    Objects of this class correspond to an SIOSoar object that was created on
    the simside.  Calls made on the SIOSoar object on the simside are encoded and 
    passed over the socket and invoke the corresponding member functions on
    the SIOSoarProxy object.
    
    @author Brad Jones
*/

class SIOSoarProxy {
public:
    /*
        This constructs an SIOSoarProxy object

        @param client the socket of communication between the SIOSoar object and
               this object.  The SIOSoarProxy now owns the socket and 
               takes responsibility for cleaning up the connection.  The socket
               should be allocated with new.
    */
    SIOSoarProxy(std::auto_ptr<sgio::Socket> client);

    
    ~SIOSoarProxy();
   
    /*
        This sends a message to the SIOSoar object sitting on the simside

        @param msg the message you want to send.  Note, the msg must be allocated
               with new.
    */
    void PostNetMessage(std::auto_ptr<sgio::Message> msg);

    /*
        This tests if a tcl command is waiting in tcl command queue.

        @return false if the tcl command queue is empty true otherwise.
    */
    bool IsATclCommandPending();

    /**
        Gets a tcl command if the tcl command queue is not empty.

        @return a command off the command queue or 0 if the queue is empty.
                note, the tclcommand should have been allocated with new, it
                is your responsibility to delete it.
    */
    std::auto_ptr<TclCommand> GetTclCommand();

    /*
        This adds a command to the tcl command queue.  

        @param tclCommand the tcl command that should eventually be executed when
               tcl gets a chance.  The tcl command should be allocated with new        
    */
    void PostTclCommand(std::auto_ptr<TclCommand> tclCommand);
    
    /*
        This runs an agent command, that is something that manipulates
        an SIOAgentProxy.  It locks the agent list so that no other thread can
        talk to the agent.

        @param inAgentName the name of the agent that the command should be run for

        @param agentCmd the agent command that should be run on the agent
    */
    void RunAgentCommand(const std::string& inAgentName,AgentCommand& agentCmd);
    
    /*
        Creates an agent with specified name

        @param inAgentName the name for the new agent, it should be unique
    */
    void CreateAgent(const std::string& inAgentName);

    /*
        Destroys an agent with the specified name

        @param inAgentName the name of the agent that should be destroyed
    */
    void DestroyAgent(const std::string& inAgentName);

    /*
        This returns a name of any agent that this SIOSoarProxy is aware of

        @return a name of any agent
    */
    std::string GetAnyAgentName();


    /*
        This method means that someone wants this object to be
        no more but doesn't own it.
    */
	void Quit() {
		Thread::Monitor m(&d_quit);
		quit = true;
	}

    /**
        Gets the status of the quit variable

        @return true if some wants this object to quit being but doesn't own it
                false otherwise
    */
	bool GetQuit() {
		Thread::Monitor m(&d_quit);
		return quit;
	}
private:
    // NOT IMPLEMENTED
    SIOSoarProxy();
    SIOSoarProxy(const SIOSoarProxy&);
    SIOSoarProxy& operator=(const SIOSoarProxy&);

    // Data Members
    // the socket we are communicating with
    std::auto_ptr<sgio::Socket> d_socket;

    // a helper object that handles send messages over the socket
    std::auto_ptr<Sender>       d_sender;

    // a helper object that handles receiving messages over the socket
    std::auto_ptr<Receiver>     d_receiver;
    
    // the lock for the agents list and the agent list
    typedef std::list<SIOAgentProxy*> AgentsList;
    Thread::Mutex        d_agentsListLock;
    AgentsList           d_agentsList;
    

    // the lock for the quit variable and the quit variable
	Thread::Mutex d_quit;
	bool quit;

    // the lock for the tcl command queue and the tcl command queue
    typedef std::list<TclCommand*> CmdQueue;
	Thread::Mutex commandsLock;
	CmdQueue commands;
};

#endif // INCLUDED_SIO_SOAR_PROXY_H