#ifndef INCLUDED_TCL_COMMANDS_H
#define INCLUDED_TCL_COMMANDS_H

#include <string>
#include <list>
#include "thread.h"

struct Tcl_Interp;
class Sender;
class Receiver;
class RunTilOutputTclCmd;
class RunAgentTilOutputTclCmd;
class StepTclCmd;
class StepAgentTclCmd;

/*
	TclCommand is an abstract base class for command objects to
	encapsulate running particular commands through a tcl interpreter
*/

class TclCommand {
public:
	TclCommand(bool inUnique = false) : unique(inUnique) {}
	virtual ~TclCommand();

    /*
        Unique is a property of a command that means
        This command could be obliterated by another command if
        it is unique
     */
	bool Unique() const { return unique; }

    /*
        Given a unique command as a parameter
        this method returns whether or not this
        command makes the other command irrelevant
    */
	virtual bool Obliterates(TclCommand* tclCommand);

	virtual void Execute(Tcl_Interp* interp) = 0;
	virtual bool IsObliteratedBy(RunTilOutputTclCmd* rtotc);
	virtual bool IsObliteratedBy(RunAgentTilOutputTclCmd* ratotc);
	virtual bool IsObliteratedBy(StepTclCmd* stc);
	virtual bool IsObliteratedBy(StepAgentTclCmd* satc);
protected:
	bool unique;
};


/**
	This class encapsulates the tcl commands that must be executed to
	add an agent to soar
*/
class AddAgentTclCmd : public TclCommand {
public:
	AddAgentTclCmd(const std::string& inAgentName) 
	: agentName(inAgentName) {}
	virtual ~AddAgentTclCmd();

	virtual void Execute(Tcl_Interp* interp);
private:
	std::string agentName;
};

/**
	This class encapsulates the tcl commands that must be executed
	to remove an agent from soar
*/
class RemoveAgentTclCmd : public TclCommand {
public:
	RemoveAgentTclCmd(const std::string& inAgentName) 
	: agentName(inAgentName) {}

	virtual ~RemoveAgentTclCmd();
	virtual void Execute(Tcl_Interp* interp);
private:
	std::string agentName;
};

/**
	LoadProductions encapsulates the code necessary source
	productions into an agent
*/
class LoadProductionsTclCmd : public TclCommand {
public:
	LoadProductionsTclCmd(const std::string& inAgentName,const std::string& inPath)
	: agentName(inAgentName), path(inPath) {}
	virtual ~LoadProductionsTclCmd();

	virtual void Execute(Tcl_Interp* interp);
private:
	std::string agentName;
	std::string path;
};

/**
	LoadReteNet encapsulates the code necessary to load a retenet
*/
class LoadReteNetTclCmd : public TclCommand {
public:
	LoadReteNetTclCmd(const std::string& inAgentName,const std::string& inPath) 
	: agentName(inAgentName), path(inPath) {}
	virtual ~LoadReteNetTclCmd();

	virtual void Execute(Tcl_Interp* interp);
private:
	std::string agentName;
	std::string path;
};


/**
	SaveReteNet encapsulates the code necessary to save a retenet
*/
class SaveReteNetTclCmd : public TclCommand {
public:
	SaveReteNetTclCmd(const std::string& inAgentName,const std::string& inPath) 
	: agentName(inAgentName), path(inPath) {}
	virtual ~SaveReteNetTclCmd();

	virtual void Execute(Tcl_Interp* interp);
private:
	std::string agentName;
	std::string path;
};

/**
	Run encapsulates the code necessary to run soar
*/
class RunTilOutputTclCmd : public TclCommand {
public:
     RunTilOutputTclCmd(const std::string& anyAgentName,int inMaxDecisions) 
     : TclCommand(true), agentName(anyAgentName),maxDecisions(inMaxDecisions) {}
	virtual ~RunTilOutputTclCmd();

	virtual bool Obliterates(TclCommand* tclCommand);
	virtual bool IsObliteratedBy(RunTilOutputTclCmd* rtotc);
	virtual void Execute(Tcl_Interp* interp);
private:
    std::string agentName;
    int maxDecisions;
};

class StepTclCmd : public TclCommand {
public:
     StepTclCmd(const std::string& anyAgentName) 
     : TclCommand(true), agentName(anyAgentName) {}
	virtual ~StepTclCmd();

	virtual bool Obliterates(TclCommand* tclCommand);
	virtual bool IsObliteratedBy(StepTclCmd* rtotc);
	virtual void Execute(Tcl_Interp* interp);
private:
    std::string agentName;
};

/**
    This class of object changes the watch level
*/
class SetWatchLevelTclCmd : public TclCommand {
public:
	SetWatchLevelTclCmd(const std::string& inAgentName,int inWatchLevel)
		: agentName(inAgentName),watchLevel(inWatchLevel) {}
	virtual ~SetWatchLevelTclCmd();

	virtual void Execute(Tcl_Interp* interp);
private:
	std::string agentName;
    int watchLevel;
};

/**
    This class of object runs a particular agent
    for 15 decision cycles or until it generates output
*/
class RunAgentTilOutputTclCmd : public TclCommand {
public:
    RunAgentTilOutputTclCmd(const std::string& agentName)
        : TclCommand(true), d_agentName(agentName) {}
    virtual ~RunAgentTilOutputTclCmd();

	virtual bool Obliterates(TclCommand* tclCommand);
	virtual bool IsObliteratedBy(RunAgentTilOutputTclCmd* ratotc);
    virtual void Execute(Tcl_Interp* interp);
private:
    std::string d_agentName; 
};

class StepAgentTclCmd : public TclCommand {
public:
    StepAgentTclCmd(const std::string& agentName)
        : TclCommand(true), d_agentName(agentName) {}
    virtual ~StepAgentTclCmd();

	virtual bool Obliterates(TclCommand* tclCommand);
	virtual bool IsObliteratedBy(StepAgentTclCmd* satc);
    virtual void Execute(Tcl_Interp* interp);
private:
    std::string d_agentName; 
};

#endif // INCLUDED_TCL_COMMANDS_H
